/*
 * cfin - find which file is being used for a given command
 *
 * James O'Toole <james@maryland>
 * Hacked to handle "~" notation by Steve Miller <steve@maryland>
 * Further hacked to handle regular expressions by Steve Miller.
 */

#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <pwd.h>

#define EXECPERM	(S_IEXEC | (S_IEXEC >> 3) | (S_IEXEC >> 6))

extern int errno;
extern char *sys_errlist[];
char **argp;
int pflag = 0;		/* toggle pattern-matching */

main(argc, argv)
	int argc;
	char **argv;
{
	int i;

	if (argc < 2) {
		fprintf(stderr, "Usage: %s [-p] command...\n", *argv);
		exit(1);
	}
	argp = argv;
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-p") == 0) {
			if (pflag)
				pflag = 0;
			else
				pflag = 1;
			continue;
		}
		doittoit(argv[i]);
	}
	exit(0);
}

doittoit(nam)
	char *nam;
{
	register char *path, *dp, *rep;
	char *getenv(), *re_comp();

	if (pflag)
		if ((rep = re_comp(nam)) != 0) {
			fprintf(stderr, "%s: %s: invalid expression (%s)\n",
			    argp[0], nam, rep);
			return;
		}

	dp = path = getenv("PATH");

	if (path == NULL) {
		dsearch(".", nam);
		return;
	}
	while (*path) {
		if (*path == ':') {
			*path = '\0';
			dsearch(dp, nam);
			*path++ = ':';
			dp = path;
		} else
			path++;
	}
	dsearch(dp, nam);
}

dsearch(dirp, cmd)
	register char *dirp, *cmd;
{
	register DIR *cd;
	register struct direct *cname;
	/* extreme */
	char dirnam[MAXPATHLEN], fullnam[MAXPATHLEN], uname[MAXPATHLEN];
	struct stat st;
	struct passwd *pwd;

	/*
	 * yucko!
	 */
	if (*dirp == '\0')	/* kludge */
		dirp = ".";
	if (*dirp == '~') {
		if ((dirp[1] == '/') || (dirp[1] == '\0')) {	/* home dir */
			if ((pwd = getpwuid(getuid())) == 0)
				return;		/* strange */
			sprintf(dirnam, "%s%s", pwd->pw_dir, ++dirp);
		} else {			/* ~something */
			int i = 1;

			while ((i - 1 < MAXPATHLEN) && (dirp[i]) &&
			    (dirp[i] != '/')) {
				uname[i - 1] = dirp[i];
				i++;
			}
			uname[i - 1] = '\0';
			if ((pwd = getpwnam(uname)) == 0)
				return;
			sprintf(dirnam, "%s%s", pwd->pw_dir, &(dirp[i]));
		}
	} else
		(void) strcpy(dirnam, dirp);

	/*
	 * Now we do the real work.  Dirnam contains the name of
	 * the directory we're currently searching; we open the directory
	 * and readdir() to the end of it, checking the name of the
	 * files in that directory against the regular expression
	 * we got from the user (in cmd).  By this time, the user's
	 * regular expression has already been compiled, so all we
	 * have to do is compare names got by readdir()s to that
	 * expression.  If we find a match, we stat() it and check
	 * permissions.
	 */
	if ((cd = opendir(dirnam)) == (DIR *)0)
		return;
	for (cname = readdir(cd) ; cname ; cname = readdir(cd)) {
		if ((pflag && re_exec(cname->d_name)) ||
		    (strcmp(cname->d_name, cmd) == 0)) {
			sprintf(fullnam, "%s/%s", dirnam, cname->d_name);
			if (stat(fullnam, &st)) {
				fprintf(stderr, "%s: can't stat %s (%s)??\n",
				   argp[0], fullnam, sys_errlist[errno]);
				continue;
			}
			if (st.st_mode & EXECPERM &&
			    (st.st_mode & S_IFMT) == S_IFREG)
				printf("%s\n", fullnam);
			else
				printf("%s: not executable\n", fullnam);
		}
	}
	(void) closedir(cd);
}
